@@ -2,13 +2,15 @@
 
 from libifstate.parser import YamlParser
 from libifstate import __version__, IfState
-from libifstate.exception import FeatureMissingError, LinkNoConfigFound, ParserValidationError, ParserOpenError, ParserIncludeError
+from libifstate.exception import FeatureMissingError, LinkNoConfigFound, LinkCircularLinked, ParserValidationError, ParserOpenError, ParserParseError, ParserIncludeError
 from libifstate.util import logger, IfStateLogging
-from collections import namedtuple
-from copy import deepcopy
+from setproctitle import setproctitle
 
 import argparse
+import glob
 import logging
+import os
+from pathlib import Path
 import re
 import signal
 import sys
@@ -34,6 +36,55 @@ ACTIONS_HELP = {
     "SHELL"    : "launch interactive python shell (pyroute2)",
 }
 
+class IfsConfigHandler():
+    def __init__(self, fn, soft_schema):
+        self.fn = fn
+        self.soft_schema = soft_schema
+
+        self.ifs = self.load_config()
+        self.cb = None
+
+    def set_callback(self, cb):
+        self.cb = cb
+
+    def sighup_handler(self, signum, frame):
+        logger.info("SIGHUP: reloading configuration")
+        try:
+            self.ifs = self.load_config()
+            self.cb(self.ifs)
+        except:
+            logger.exception("failed to reload configuration")
+
+    def load_config(self):
+        try:
+            parser = YamlParser(self.fn)
+        except ParserOpenError as ex:
+            logger.error(
+                "Config loading from {} failed: {}".format(ex.fn, ex.msg))
+            raise ex
+        except ParserParseError as ex:
+            logger.error("Config parsing failed:\n\n{}".format(str(ex)))
+            raise ex
+        except ParserIncludeError as ex:
+            logger.error(
+                "Config include file {} failed: {}".format(ex.fn, ex.msg))
+            raise ex
+
+        try:
+            ifstates = parser.config()
+
+            ifs = IfState()
+            ifs.update(ifstates, self.soft_schema)
+            return ifs
+        except ParserValidationError as ex:
+            logger.error("Config validation failed for {}".format(ex.detail))
+            raise ex
+        except FeatureMissingError as ex:
+            logger.error(
+                "Config uses unavailable feature: {}".format(ex.feature))
+            raise ex
+
+
 def shell():
     from ifstate.shell import IfStateConsole
     import pyroute2
@@ -42,6 +93,29 @@ def shell():
     shell.interact(banner=f"ifstate {__version__}; pyroute2 {pyroute2.__version__}")
 
 
+def sighup_vrrp_fifo():
+    prefix = 'ifstate-vrrp-fifo@'
+    pids = {}
+
+    for pid_file in glob.glob('/run/libifstate/vrrp/[1-9]*.pid'):
+        pid = Path(pid_file).stem
+        if pid.isnumeric():
+            try:
+                with open(f'/proc/{pid}/cmdline', encoding='utf-8') as fh:
+                    if prefix == fh.read(len(prefix)):
+                        # send SIGHUP to reload config
+                        os.kill(int(pid), signal.SIGHUP)
+
+                        pids[pid] = fh.read(32)
+            except OSError:
+                pass
+
+    if len(pids):
+        logger.info("")
+        logger.info("notified vrrp-fifo handlers...")
+        for pid, fifo in sorted(pids.items()):
+            logger.info(fifo, extra={'option': pid, 'style': IfStateLogging.STYLE_CHG})
+
 def main():
     parser = argparse.ArgumentParser()
     group = parser.add_mutually_exclusive_group()
@@ -82,18 +156,23 @@ def main():
     else:
         lvl = logging.INFO
 
+    if args.action == Actions.VRRP_FIFO:
+        setproctitle("ifstate-{}@{}".format(args.action, args.fifo))
+    else:
+        setproctitle("ifstate-{}".format(args.action))
+
     if args.action == Actions.SHELL:
         shell()
         exit(0)
 
-    ifslog = IfStateLogging(lvl)
-    ifs = IfState()
+    ifslog = IfStateLogging(lvl, action=args.action)
 
     if args.action in [Actions.SHOW, Actions.SHOWALL]:
         # preserve dict order on python 3.7+
         if sys.version_info >= (3, 7):
             yaml.add_representer(
                 dict, lambda self, data: yaml.representer.SafeRepresenter.represent_dict(self, data.items()))
+        ifs = IfState()
         print(yaml.dump(ifs.show(args.action == Actions.SHOWALL)))
 
         ifslog.quit()
@@ -101,56 +180,26 @@ def main():
 
     if args.action in [Actions.CHECK, Actions.APPLY, Actions.VRRP, Actions.VRRP_FIFO]:
         try:
-            parser = YamlParser(args.config)
-        except ParserOpenError as ex:
-            logger.error(
-                "Config loading from {} failed: {}".format(ex.fn, ex.msg))
+            ifs_config = IfsConfigHandler(args.config, args.soft_schema)
+        except (ParserOpenError,
+                ParserParseError,
+                ParserIncludeError,
+                ParserValidationError,
+                FeatureMissingError) as ex:
             ifslog.quit()
-            exit(1)
-        except yaml.parser.ParserError as ex:
-            logger.error("Config parsing failed:\n\n{}".format(str(ex)))
-            ifslog.quit()
-            exit(2)
-        except ParserIncludeError as ex:
-            logger.error(
-                "Config include file {} failed: {}".format(ex.fn, ex.msg))
-            ifslog.quit()
-            exit(3)
-
-        try:
-            ifstates = parser.config()
-            ifs.update(ifstates, args.soft_schema)
-        except ParserValidationError as ex:
-            logger.error("Config validation failed for {}".format(ex.detail))
-            ifslog.quit()
-            exit(4)
-        except FeatureMissingError as ex:
-            logger.error(
-                "Config uses unavailable feature: {}".format(ex.feature))
-            ifslog.quit()
-            exit(5)
+            exit(ex.exit_code())
 
         if args.action == Actions.CHECK:
             try:
-                ifs.check()
+                ifs_config.ifs.check()
             except LinkNoConfigFound:
                 pass
+            except LinkCircularLinked as ex:
+                ifslog.quit()
+                exit(ex.exit_code())
         elif args.action == Actions.VRRP_FIFO:
-            status_pattern = re.compile(
-                r'(group|instance) "([^"]+)" (unknown|fault|backup|master)$', re.IGNORECASE)
-
-            with open(args.fifo) as fifo:
-                for line in fifo:
-                    m = status_pattern.match(line.strip())
-                    if m:
-                        signal.signal(signal.SIGHUP, signal.SIG_IGN)
-                        signal.signal(signal.SIGPIPE, signal.SIG_IGN)
-                        signal.signal(signal.SIGTERM, signal.SIG_IGN)
-                        try:
-                            ifs_tmp = deepcopy(ifs)
-                            ifs_tmp.apply(m.group(1), m.group(2), m.group(3))
-                        except:
-                            pass
+            from ifstate.vrrp import vrrp_fifo
+            vrrp_fifo(args.fifo, ifs_config, lvl)
         else:
             # ignore some well-known signals to prevent interruptions (i.e. due to ssh connection loss)
             signal.signal(signal.SIGHUP, signal.SIG_IGN)
@@ -158,12 +207,16 @@ def main():
             signal.signal(signal.SIGTERM, signal.SIG_IGN)
             try:
                 if args.action == Actions.APPLY:
-                    ifs.apply()
+                    ifs_config.ifs.apply()
+                    sighup_vrrp_fifo()
                 elif args.action == Actions.VRRP:
-                    ifs.apply(
+                    ifs_config.ifs.apply(
                         args.type, args.name, args.state)
             except LinkNoConfigFound:
                 pass
+            except LinkCircularLinked as ex:
+                ifslog.quit()
+                exit(ex.exit_code())
 
         ifslog.quit()
 
