@@ -1,9 +1,22 @@
 from libifstate.util import logger
 from libifstate.exception import ParserValidationError
 from abc import ABC, abstractmethod
+from copy import deepcopy
 
 
 class Parser(ABC):
+    _default_lo_link = {
+        'name': 'lo',
+        'addresses': [
+            '127.0.0.1/8',
+            '::1/128',
+        ],
+        'link': {
+            'kind': 'physical',
+            'state': 'up',
+            'mtu': 65536,
+        }
+    }
     _default_ifstates = {
         'ignore': {
             'ipaddr_builtin': [
@@ -13,12 +26,15 @@ class Parser(ABC):
             'ifname_builtin': [
                 r'^br-[\da-f]{12}',
                 r'^docker\d+',
-                r'^lo$',
                 r'^ppp\d+$',
                 r'^veth',
                 r'^virbr\d+',
                 r'^vrrp\d*\.\d+$'
             ],
+            'fdb_builtin': [
+                r'^33:33:',
+                r'^01:00:5e:'
+            ],
             'routes_builtin': [
                 {'proto': 1},
                 {'proto': 2},
@@ -96,9 +112,17 @@ class Parser(ABC):
                     a[key] = b[key]
         return a
 
+    def _update_lo(self, cfg):
+        if 'interfaces' in cfg:
+            lo = next((idx for idx, iface in enumerate(cfg['interfaces']) if iface['name'] == 'lo'), None)
+            if lo is not None:
+                cfg['interfaces'][lo] = self.merge(deepcopy(Parser._default_lo_link), cfg['interfaces'][lo])
+            else:
+                cfg['interfaces'].append(Parser._default_lo_link)
+
     def config(self):
         # merge builtin defaults with config
-        cfg = (self.merge(self._default_ifstates, self.ifstates))
+        cfg = (self.merge(deepcopy(Parser._default_ifstates), self.ifstates))
 
         # 'ignore' should still be an object
         try:
@@ -106,6 +130,11 @@ class Parser(ABC):
         except TypeError:
             raise ParserValidationError("$.ignore: is not of type 'object'")
 
+        # add loopback interface defaults
+        self._update_lo(cfg)
+        for namespace in cfg.get('namespaces', {}):
+            self._update_lo(cfg['namespaces'][namespace])
+
         # merge builtin defaults
         for k in list(cfg["ignore"]):
             if k.endswith("_builtin"):
