@@ -1,4 +1,4 @@
-from libifstate.util import logger, ipr, IfStateLogging
+from libifstate.util import logger, IfStateLogging
 from libifstate.exception import RouteDupblicate, netlinkerror_classes
 from ipaddress import ip_address, ip_network, IPv6Network
 from pyroute2.netlink.rtnl.fibmsg import FR_ACT_VALUES
@@ -19,7 +19,6 @@ def route_matches(r1, r2, fields=('dst', 'priority', 'proto'), indent=None):
 def rule_matches(r1, r2, fields=('priority', 'iif', 'oif', 'dst', 'metric', 'protocol'), indent=None):
     return _matches(r1, r2, fields, indent)
 
-
 def _matches(r1, r2, fields, indent):
     for fld in fields:
         if not indent is None:
@@ -31,6 +30,30 @@ def _matches(r1, r2, fields, indent):
             return False
     return True
 
+VRRP_MATCH_IGNORE = 0
+VRRP_MATCH_DISABLE = 1
+VRRP_MATCH_ENABLE =  2
+
+def vrrp_match(obj, by_vrrp, vrrp_type, vrrp_name, vrrp_state):
+    if by_vrrp:
+        # test if object has a vrrp condition at all
+        if not '_vrrp' in obj:
+            return VRRP_MATCH_IGNORE
+
+        # test if vrrp type and name matches
+        if vrrp_type != obj['_vrrp']['type'] or vrrp_name != obj['_vrrp']['name']:
+            return VRRP_MATCH_IGNORE
+
+        # test if state is matching
+        if vrrp_state in obj['_vrrp']['states']:
+            return VRRP_MATCH_ENABLE
+        else:
+            return VRRP_MATCH_DISABLE
+    else:
+        # ignore vrrp object when not in vrrp action
+        if '_vrrp' in obj:
+            return VRRP_MATCH_IGNORE
+
 
 class RTLookup():
     def __init__(self, name):
@@ -38,16 +61,20 @@ class RTLookup():
         self.str2id = {}
         self.id2str = {}
 
-        try:
-            fn = os.path.join('/etc/iproute2', name)
-            with open(fn, 'r') as fp:
-                self._parse(fp)
-
-            for fn in glob(os.path.join('/etc/iproute2', "{}.d".format(name), "*.conf")):
+        for basedir in ['/usr/lib/iproute2', '/etc/iproute2']:
+            fn = os.path.join(basedir, name)
+            try:
                 with open(fn, 'r') as fp:
                     self._parse(fp)
-        except:
-            logger.info('could not open {}'.format(fn))
+            except IOError as err:
+                pass
+
+        for fn in glob(os.path.join('/etc/iproute2', "{}.d".format(name), "*.conf")):
+            try:
+                with open(fn, 'r') as fp:
+                    self._parse(fp)
+            except IOError as err:
+                pass
 
     def _parse(self, fp):
         for line in fp:
@@ -89,11 +116,11 @@ RT_LOOKUPS_DEFAULTS = {
 }
 
 class Tables(collections.abc.Mapping):
-    def __init__(self):
+    def __init__(self, netns):
+        self.netns = netns
         self.tables = {
             254: [],
         }
-
     def __getitem__(self, key):
         if not key in self.tables:
             raise KeyError()
@@ -120,7 +147,7 @@ class Tables(collections.abc.Mapping):
             except KeyError as err:
                 # mapping not available - catch exception and skip it
                 logger.warning('ignoring unknown %s "%s"', key, rt[key],
-                               extra={'iface': rt['dst']})
+                               extra={'iface': rt['dst'], 'netns': self.netns})
                 rt[key] = RT_LOOKUPS_DEFAULTS[key]
 
         if type(rt['type']) == str:
@@ -142,13 +169,16 @@ class Tables(collections.abc.Mapping):
         else:
             rt['priority'] = 0
 
+        if 'vrrp' in route:
+            rt['_vrrp'] = route['vrrp']
+
         if not rt['table'] in self.tables:
             self.tables[rt['table']] = []
         self.tables[rt['table']].append(rt)
 
     def show_routes(self, ignores):
         routes = []
-        for route in ipr.get_routes(family=AF_INET) + ipr.get_routes(family=AF_INET6):
+        for route in self.netns.ipr.get_routes(family=AF_INET) + self.netns.ipr.get_routes(family=AF_INET6):
             # skip routes from local table
             table = route.get_attr('RTA_TABLE')
             if table == 255:
@@ -180,7 +210,7 @@ class Tables(collections.abc.Mapping):
 
             dev = route.get_attr('RTA_OIF')
             if dev:
-                link = next(iter(ipr.get_links(dev)), None)
+                link = next(iter(self.netns.ipr.get_links(dev)), None)
                 if link:
                     rt['dev'] = link.get_attr('IFLA_IFNAME', dev)
                 else:
@@ -221,7 +251,7 @@ class Tables(collections.abc.Mapping):
 
     def kernel_routes(self, table):
         routes = []
-        for route in ipr.get_routes(table=table, family=AF_INET) + ipr.get_routes(table=table, family=AF_INET6):
+        for route in self.netns.ipr.get_routes(table=table, family=AF_INET) + self.netns.ipr.get_routes(table=table, family=AF_INET6):
             # ignore RTM_F_CLONED routes
             if route['flags'] & 512:
                 continue
@@ -269,48 +299,62 @@ class Tables(collections.abc.Mapping):
             routes.append(rt)
         return routes
 
-    def apply(self, ignores, do_apply):
+    def apply(self, ignores, do_apply, by_vrrp, vrrp_type, vrrp_name, vrrp_state):
         for table, croutes in self.tables.items():
-            pfx = RTLookups.tables.lookup_str(table)
-            logger.info('\nconfiguring routing table {}...'.format(pfx))
+            log_str = RTLookups.tables.lookup_str(table)
+            if self.netns.netns != None:
+                log_str += "[netns={}]".format(self.netns.netns)
 
             kroutes = self.kernel_routes(table)
 
             for route in croutes:
                 if 'oif' in route and type(route['oif']) == str:
                     route['oif'] = next(
-                        iter(ipr.link_lookup(ifname=route['oif'])), None)
+                        iter(self.netns.ipr.link_lookup(ifname=route['oif'])), None)
                 found = False
                 identical = False
+                matched = vrrp_match(route, by_vrrp, vrrp_type, vrrp_name, vrrp_state)
                 for i, kroute in enumerate(kroutes):
                     if route_matches(route, kroute):
+                        # ignore kernel routes due to vrrp_match
+                        if matched == VRRP_MATCH_IGNORE:
+                            del kroutes[i]
+                            break
+
+                        # remove kernel routes due to vrrp_match
+                        if matched == VRRP_MATCH_DISABLE:
+                            break
+
                         del kroutes[i]
                         found = True
-                        if route_matches(route, kroute, route.keys(), indent=route['dst']):
+                        if route_matches(
+                            route,
+                            kroute,
+                            [key for key in route.keys() if key[0] != '_'],
+                            indent=route['dst']
+                        ):
                             identical = True
                             break
 
-                if identical:
-                    logger.info(
-                        'ok', extra={'iface': route['dst'], 'style': IfStateLogging.STYLE_OK})
-                else:
-                    if found:
-                        logger.info('change', extra={
-                                    'iface': route['dst'], 'style': IfStateLogging.STYLE_CHG})
+                if matched not in [VRRP_MATCH_IGNORE, VRRP_MATCH_DISABLE]:
+                    if identical:
+                        logger.log_ok(log_str, "= {}".format(route['dst']))
                     else:
-                        logger.info(
-                            'add', extra={'iface': route['dst'], 'style': IfStateLogging.STYLE_CHG})
+                        if found:
+                            logger.log_change(log_str, "~ {}".format(route['dst']))
+                        else:
+                            logger.log_add(log_str, "+ {}".format(route['dst']))
 
-                    logger.debug("ip route replace: {}".format(
-                        " ".join("{}={}".format(k, v) for k, v in route.items())))
-                    try:
-                        if do_apply:
-                            ipr.route('replace', **route)
-                    except Exception as err:
-                        if not isinstance(err, netlinkerror_classes):
-                            raise
-                        logger.warning('route setup {} failed: {}'.format(
-                            route['dst'], err.args[1]))
+                        logger.debug("ip route replace: {}".format(
+                            " ".join("{}={}".format(k, v) for k, v in route.items())))
+                        try:
+                            if do_apply:
+                                self.netns.ipr.route('replace', **route)
+                        except Exception as err:
+                            if not isinstance(err, netlinkerror_classes):
+                                raise
+                            logger.warning('route setup {} failed: {}'.format(
+                                route['dst'], err.args[1]))
 
             for route in kroutes:
                 ignore = False
@@ -321,11 +365,10 @@ class Tables(collections.abc.Mapping):
                 if ignore:
                     continue
 
-                logger.info(
-                    'del', extra={'iface': route['dst'], 'style': IfStateLogging.STYLE_DEL})
+                logger.log_del(log_str, "- {}".format(route['dst']))
                 try:
                     if do_apply:
-                        ipr.route('del', **route)
+                        self.netns.ipr.route('del', **route)
                 except Exception as err:
                     if not isinstance(err, netlinkerror_classes):
                         raise
@@ -334,7 +377,8 @@ class Tables(collections.abc.Mapping):
 
 
 class Rules():
-    def __init__(self):
+    def __init__(self, netns):
+        self.netns = netns
         self.rules = []
 
     def add(self, rule):
@@ -382,11 +426,14 @@ class Rules():
         if 'priority' in rule:
             ru['priority'] = rule['priority']
 
+        if 'vrrp' in rule:
+            ru['_vrrp'] = rule['vrrp']
+
         self.rules.append(ru)
 
     def kernel_rules(self):
         rules = []
-        for rule in ipr.get_rules(family=AF_INET) + ipr.get_rules(family=AF_INET6):
+        for rule in self.netns.ipr.get_rules(family=AF_INET) + self.netns.ipr.get_rules(family=AF_INET6):
             ru = {
                 'action': FR_ACT_VALUES.get(rule['action']),
                 'table': rule.get_attr('FRA_TABLE'),
@@ -455,33 +502,45 @@ class Rules():
 
         return rules
 
-    def apply(self, ignores, do_apply):
-        logger.info('\nconfiguring routing rules...')
+    def apply(self, ignores, do_apply, by_vrrp, vrrp_type, vrrp_name, vrrp_state):
         krules = self.kernel_rules()
         for rule in self.rules:
+            log_str = '#{}'.format(rule['priority'])
+            if self.netns.netns != None:
+                log_str += "[netns={}]".format(self.netns.netns)
+
             found = False
+            matched = vrrp_match(rule, by_vrrp, vrrp_type, vrrp_name, vrrp_state)
             for i, krule in enumerate(krules):
-                if rule_matches(rule, krule, rule.keys()):
-                    del krules[i]
+                if rule_matches(
+                    rule,
+                    krule,
+                    # skip helper attrs like _vrrp
+                    [key for key in rule.keys() if key[0] != '_']
+                ):
                     found = True
+
+                    # remove kernel routes due to vrrp_match
+                    if matched != VRRP_MATCH_DISABLE:
+                        del krules[i]
+
                     break
 
-            if found:
-                logger.info(
-                    'ok', extra={'iface': '#{}'.format(rule['priority']), 'style': IfStateLogging.STYLE_OK})
-            else:
-                logger.info(
-                    'add', extra={'iface': '#{}'.format(rule['priority']), 'style': IfStateLogging.STYLE_CHG})
+            if matched not in [VRRP_MATCH_IGNORE, VRRP_MATCH_DISABLE]:
+                if found:
+                    logger.log_ok(log_str)
+                else:
+                    logger.log_add(log_str)
 
-                logger.debug("ip rule add: {}".format(
-                    " ".join("{}={}".format(k, v) for k, v in rule.items())))
-                try:
-                    if do_apply:
-                        ipr.rule('add', **rule)
-                except Exception as err:
-                    if not isinstance(err, netlinkerror_classes):
-                        raise
-                    logger.warning('rule setup failed: {}'.format(err.args[1]))
+                    logger.debug("ip rule add: {}".format(
+                        " ".join("{}={}".format(k, v) for k, v in rule.items())))
+                    try:
+                        if do_apply:
+                            self.netns.ipr.rule('add', **rule)
+                    except Exception as err:
+                        if not isinstance(err, netlinkerror_classes):
+                            raise
+                        logger.warning('rule setup failed: {}'.format(err.args[1]))
 
         for rule in krules:
             ignore = False
@@ -495,11 +554,10 @@ class Rules():
             if ignore:
                 continue
 
-            logger.info(
-                'del', extra={'iface': '#{}'.format(rule['priority']), 'style': IfStateLogging.STYLE_DEL})
+            logger.log_del(log_str)
             try:
                 if do_apply:
-                    ipr.rule('del', **rule)
+                    self.netns.ipr.rule('del', **rule)
             except Exception as err:
                 if not isinstance(err, netlinkerror_classes):
                     raise
