@@ -1,6 +1,8 @@
-from libifstate.util import logger, ipr, IfStateLogging
+from libifstate.util import logger, IfStateLogging
 from libifstate.exception import ExceptionCollector, netlinkerror_classes
 
+import errno
+from pyroute2 import NetlinkError
 
 class TC():
     ROOT_HANDLE = 0xFFFFFFFF
@@ -33,18 +35,19 @@ class TC():
 
         return int(l[0], 16) << 16 | int(l[1], 16)
 
-    def __init__(self, iface, tc):
+    def __init__(self, netns, iface, tc):
+        self.netns = netns
         self.iface = iface
         self.idx = None
         self.tc = tc
 
-    def get_qdisc(self, ipr_qdiscs, parent):
-        for qdisc in ipr_qdiscs:
+    def get_qdisc(self, netns_qdiscs, parent):
+        for qdisc in netns_qdiscs:
             if qdisc['parent'] == parent:
                 return qdisc
 
-    def get_qchild(self, ipr_qdiscs, parent, slot):
-        for qdisc in ipr_qdiscs:
+    def get_qchild(self, netns_qdiscs, parent, slot):
+        for qdisc in netns_qdiscs:
             if qdisc['parent'] == parent | slot:
                 return qdisc
 
@@ -58,7 +61,7 @@ class TC():
                         "parent": TC.INGRESS_PARENT,
                     }
                     try:
-                        ipr.tc("del", **opts)
+                        self.netns.ipr.tc("del", **opts)
                     except Exception as err:
                         if not isinstance(err, netlinkerror_classes):
                             raise
@@ -74,7 +77,7 @@ class TC():
                         "kind": "ingress",
                     }
                     try:
-                        ipr.tc("add", **opts)
+                        self.netns.ipr.tc("add", **opts)
                     except Exception as err:
                         if not isinstance(err, netlinkerror_classes):
                             raise
@@ -86,6 +89,20 @@ class TC():
         return False
 
     def apply_qtree(self, tc, qdisc, ipr_qdiscs, parent, excpts, do_apply, recreate=False):
+        # if tc is None, we shall cleanup the qdisc tree
+        # (remove the root qdisc to restore the interface's default)
+        if tc is None:
+            if do_apply:
+                try:
+                    self.netns.ipr.tc("del", index=self.idx, parent=parent)
+                    return True
+                except NetlinkError as ex:
+                    # the native qdisc cannot be removed but only replaced,
+                    # so we ignore any ENOENT
+                    if ex.code != errno.ENOENT:
+                        logger.exception('cannot remove qdisc')
+            return False
+
         if qdisc is None:
             recreate = True
 
@@ -103,9 +120,12 @@ class TC():
         if recreate and do_apply:
             if qdisc is not None:
                 try:
-                    ipr.tc("del", index=self.idx, parent=qdisc["parent"])
-                except:
-                    pass
+                    self.netns.ipr.tc("del", index=self.idx, parent=qdisc["parent"])
+                except NetlinkError as ex:
+                    # the native qdisc cannot be removed but only replaced,
+                    # so we ignore any ENOENT
+                    if ex.code != errno.ENOENT:
+                        logger.exception('cannot remove qdisc')
 
         opts = {
             "index": self.idx,
@@ -118,7 +138,7 @@ class TC():
         if do_apply:
             if recreate:
                 try:
-                    ipr.tc("add", **opts)
+                    self.netns.ipr.tc("add", **opts)
                 except Exception as err:
                     if not isinstance(err, netlinkerror_classes):
                         raise
@@ -128,7 +148,7 @@ class TC():
             else:
                 # soft update for TCA_OPTIONS
                 try:
-                    ipr.tc("change", **opts)
+                    self.netns.ipr.tc("change", **opts)
                 except Exception as err:
                     if not isinstance(err, netlinkerror_classes):
                         raise
@@ -179,7 +199,7 @@ class TC():
                         "parent": parent,
                     }
                     try:
-                        ipr.del_filter_by_info(**opts)
+                        self.netns.ipr.del_filter_by_info(**opts)
                     except Exception as err:
                         if not isinstance(err, netlinkerror_classes):
                             raise
@@ -196,7 +216,7 @@ class TC():
                             if action["kind"] == "mirred":
                                 # get ifindex
                                 action["ifindex"] = next(
-                                    iter(ipr.link_lookup(ifname=action["dev"])), None)
+                                    iter(self.netns.ipr.link_lookup(ifname=action["dev"])), None)
 
                                 if self.idx == None:
                                     logger.warning("filter #{} references unknown interface {}".format(
@@ -206,7 +226,7 @@ class TC():
                         tc_filter["parent"] = TC.handle2int(tc_filter["parent"])
                     try:
                         try:
-                            ipr.tc("replace-filter", **tc_filter)
+                            self.netns.ipr.tc("replace-filter", **tc_filter)
                             # replace seems only to work if there is no filter
                             # => something has changed
                             changes = True
@@ -221,8 +241,8 @@ class TC():
                                 "info": tc_filter["prio"] << 16,
                                 "parent": parent,
                             }
-                            ipr.del_filter_by_info(**opts)
-                            ipr.tc("add-filter", **tc_filter)
+                            self.netns.ipr.del_filter_by_info(**opts)
+                            self.netns.ipr.tc("add-filter", **tc_filter)
 
                     except Exception as err:
                         if not isinstance(err, netlinkerror_classes):
@@ -237,7 +257,7 @@ class TC():
         excpts = ExceptionCollector(ifname=self.iface)
 
         # get ifindex
-        self.idx = next(iter(ipr.link_lookup(ifname=self.iface)), None)
+        self.idx = next(iter(self.netns.ipr.link_lookup(ifname=self.iface)), None)
 
         if self.idx == None:
             logger.warning('link missing', extra={'iface': self.iface})
@@ -248,7 +268,7 @@ class TC():
 
         # apply ingress qdics
         if "ingress" in self.tc:
-            ipr_qdiscs = ipr.get_qdiscs(index=self.idx)
+            ipr_qdiscs = self.netns.ipr.get_qdiscs(index=self.idx)
             if self.apply_ingress(self.tc["ingress"],
                                   self.get_qdisc(
                                       ipr_qdiscs, TC.INGRESS_PARENT),
@@ -259,7 +279,7 @@ class TC():
         # apply qdisc tree
         if "qdisc" in self.tc:
             if ipr_qdiscs is None:
-                ipr_qdiscs = ipr.get_qdiscs(index=self.idx)
+                ipr_qdiscs = self.netns.ipr.get_qdiscs(index=self.idx)
             logger.debug('checking qdisc tree', extra={'iface': self.iface})
             if self.apply_qtree(
                     self.tc["qdisc"],
@@ -272,8 +292,8 @@ class TC():
 
         # apply filters
         if "filter" in self.tc:
-            ipr_filters = ipr.get_filters(
-                index=self.idx) + ipr.get_filters(index=self.idx, parent=TC.INGRESS_HANDLE)
+            ipr_filters = self.netns.ipr.get_filters(
+                index=self.idx) + self.netns.ipr.get_filters(index=self.idx, parent=TC.INGRESS_HANDLE)
             logger.debug('checking filters', extra={'iface': self.iface})
             if self.apply_filter(
                     self.tc["filter"],
@@ -283,12 +303,8 @@ class TC():
                 changes.append("filter")
 
         if len(changes) > 0:
-            logger.info(
-                'change ({})'.format(", ".join(changes)),
-                extra={'iface': self.iface, 'style': IfStateLogging.STYLE_CHG})
+            logger.log_change('tc', 'change ({})'.format(", ".join(changes)))
         else:
-            logger.info(
-                'ok',
-                extra={'iface': self.iface, 'style': IfStateLogging.STYLE_OK})
+            logger.log_ok('tc')
 
         return excpts
